(*----------------------------------------------------------------------------*
 *  Direct3D sample from DirectX 9.0 SDK Summer 2003                          *
 *  Delphi adaptation by Alexey Barkovoy (e-mail: clootie@ixbt.com)           *
 *                                                                            *
 *  Delphi versions 5-7 are supported                                         *
 *                                                                            *
 *  Modified: 30-Dec-2003                                                     *
 *                                                                            *
 *  Latest version can be downloaded from:                                    *
 *     http://clootie.narod.ru/delphi                                         *
 *----------------------------------------------------------------------------*)
//-----------------------------------------------------------------------------
// File: ClipMirror.cpp
//
// Desc: This sample shows how to use clip planes to implement a planar mirror.
//       The scene is reflected in a mirror and rendered in a 2nd pass. The
//       corners of the mirrors, together with the camera eye point, are used
//       to define a custom set of clip planes so that the reflected geometry
//       appears only within the mirror's boundaries.
//
//       Note: This code uses the D3D Framework helper library.
//
// Copyright (c) Microsoft Corporation. All rights reserved.
//-----------------------------------------------------------------------------

unit ClipMirrorUnit;

{$I DirectX.inc}

interface

uses
  Windows, Messages, SysUtils,
  Direct3D9, D3DX9, D3DApp, D3DEnumeration, D3DSettings,
  D3DFile, D3DFont, D3DUtil, DXUtil;


//-----------------------------------------------------------------------------
// Name: "TMirrorVertex"
// Desc: Custom mirror vertex type
//-----------------------------------------------------------------------------
type
  PMirrorVertex = ^TMirrorVertex;
  TMirrorVertex = packed record
    p: TD3DXVector3;
    n: TD3DXVector3;
    color: DWORD;
  end;
  PAMirrorVertex = ^AMirrorVertex;
  AMirrorVertex = array[0..4] of TMirrorVertex;

const
  TMirrorVertex_FVF = (D3DFVF_XYZ or D3DFVF_NORMAL or D3DFVF_DIFFUSE);

type
//-----------------------------------------------------------------------------
// Name: class CMyD3DApplication
// Desc: Application class. The base class (CD3DApplication) provides the
//       generic functionality needed in all Direct3D samples. CMyD3DApplication
//       adds functionality specific to this sample program.
//-----------------------------------------------------------------------------
  CMyD3DApplication = class(CD3DApplication)

    m_pFont: CD3DFont;           // Font for drawing text
    m_pTeapot: CD3DMesh;         // The teapot object
    m_matTeapot: TD3DXMatrixA16; // Teapot's local matrix

    m_pMirrorVB: IDirect3DVertexBuffer9;

    m_vEyePt: TD3DXVector3;      // Vectors defining the camera
    m_vLookatPt: TD3DXVector3;
    m_vUpVec: TD3DXVector3;

    function RenderMirror: HRESULT;
    function RenderScene: HRESULT;
  protected
    function OneTimeSceneInit: HRESULT; override;
    function InitDeviceObjects: HRESULT; override;
    function RestoreDeviceObjects: HRESULT; override;
    function InvalidateDeviceObjects: HRESULT; override;
    function DeleteDeviceObjects: HRESULT; override;
    function Render: HRESULT; override;
    function FrameMove: HRESULT; override;
    function ConfirmDevice(const pCaps: TD3DCaps9; dwBehavior: DWORD;
      adapterFormat: TD3DFormat; backBufferFormat: TD3DFormat): HRESULT; override;
    function FinalCleanup: HRESULT; override;
  public
    constructor Create;
  end;

implementation


//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
constructor CMyD3DApplication.Create;
begin
  inherited;
  m_strWindowTitle    := 'ClipMirror: Using D3D Clip Planes';
  m_d3dEnumeration.AppUsesDepthBuffer := True;
  m_pFont             := CD3DFont.Create('Arial', 12, D3DFONT_BOLD);
  m_pTeapot           := CD3DMesh.Create;
  m_pMirrorVB         := nil;
end;


//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
function CMyD3DApplication.OneTimeSceneInit: HRESULT;
begin
  // Initialize the camera's orientation
  m_vEyePt    := D3DXVector3(0.0, 2.0, -6.5);
  m_vLookatPt := D3DXVector3(0.0, 0.0,  0.0);
  m_vUpVec    := D3DXVector3(0.0, 1.0,  0.0);

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FrameMove: HRESULT;
var
  quat: TD3DXQuaternion;
  matView: TD3DXMatrix;
begin
  // Set the teapot's local matrix (rotating about the y-axis)
  D3DXMatrixRotationY(m_matTeapot, m_fTime);

  // When the window has focus, let the mouse adjust the camera view
  if (GetFocus <> 0) then
  begin
    quat := D3DUtil_GetRotationFromCursor(m_hWnd);
    m_vEyePt.x := 5*quat.y;
    m_vEyePt.y := 5*quat.x;
    m_vEyePt.z := -sqrt(50.0 - 25*quat.x*quat.x - 25*quat.y*quat.y);

    D3DXMatrixLookAtLH(matView, m_vEyePt, m_vLookatPt, m_vUpVec);
    m_pd3dDevice.SetTransform(D3DTS_VIEW, matView);
  end;

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: RenderScene()
// Desc: Renders all objects in the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.RenderScene: HRESULT;
var
  matLocal, matWorldSaved: TD3DXMatrix;
begin
  m_pd3dDevice.GetTransform(D3DTS_WORLD, matWorldSaved);

  // Build the local matrix
  D3DXMatrixMultiply(matLocal, m_matTeapot, matWorldSaved);
  m_pd3dDevice.SetTransform(D3DTS_WORLD, matLocal);

  // Render the object
  m_pTeapot.Render(m_pd3dDevice);

  // Restore the modified render states
  m_pd3dDevice.SetTransform(D3DTS_WORLD, matWorldSaved);

  // Output statistics
  m_pFont.DrawText(2,  0, D3DCOLOR_ARGB(255,255,255,0), m_strFrameStats);
  m_pFont.DrawText(2, 20, D3DCOLOR_ARGB(255,255,255,0), m_strDeviceStats);

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: RenderMirror()
// Desc: Renders the scene as reflected in a mirror. The corners of the mirror
//       define a plane, which is used to build the reflection matrix. The
//       scene is rendered with the cull-mode reversed, since all normals in
//       the scene are likewise reflected.
//-----------------------------------------------------------------------------
function CMyD3DApplication.RenderMirror: HRESULT;
var
  matWorldSaved: TD3DXMatrix;
  matReflectInMirror: TD3DXMatrix;
  plane: TD3DXPlane;
  a,b,c,d: TD3DXVector3;
begin
  // Save the world matrix so it can be restored
  m_pd3dDevice.GetTransform(D3DTS_WORLD, matWorldSaved);

  // Get the four corners of the mirror. (This should be dynamic rather than
  // hardcoded.)
  a:= D3DXVector3(-1.5, 1.5, 3.0);
  b:= D3DXVector3( 1.5, 1.5, 3.0);
  c:= D3DXVector3(-1.5,-1.5, 3.0);
  d:= D3DXVector3( 1.5,-1.5, 3.0);

  // Construct the reflection matrix
  D3DXPlaneFromPoints(plane, a, b, c);
  D3DXMatrixReflect(matReflectInMirror, plane);
  m_pd3dDevice.SetTransform(D3DTS_WORLD, matReflectInMirror);

  // Reverse the cull mode (since normals will be reflected)
  m_pd3dDevice.SetRenderState(D3DRS_CULLMODE, D3DCULL_CW);

  // Set the custom clip planes (so geometry is clipped by mirror edges).
  // This is the heart of this sample. The mirror has 4 edges, so there are
  // 4 clip planes, each defined by two mirror vertices and the eye point.
  m_pd3dDevice.SetClipPlane(0, PSingle(D3DXPlaneFromPoints(plane, b, a, m_vEyePt)));
  m_pd3dDevice.SetClipPlane(1, PSingle(D3DXPlaneFromPoints(plane, d, b, m_vEyePt)));
  m_pd3dDevice.SetClipPlane(2, PSingle(D3DXPlaneFromPoints(plane, c, d, m_vEyePt)));
  m_pd3dDevice.SetClipPlane(3, PSingle(D3DXPlaneFromPoints(plane, a, c, m_vEyePt)));
  m_pd3dDevice.SetRenderState(D3DRS_CLIPPLANEENABLE,
    D3DCLIPPLANE0 or D3DCLIPPLANE1 or D3DCLIPPLANE2 or D3DCLIPPLANE3);

  // Render the scene
  RenderScene;

  // Restore the modified render states
  m_pd3dDevice.SetTransform(D3DTS_WORLD, matWorldSaved);
  m_pd3dDevice.SetRenderState(D3DRS_CLIPPLANEENABLE, $00);
  m_pd3dDevice.SetRenderState(D3DRS_CULLMODE, D3DCULL_CCW);

  // Finally, render the mirror itself (as an alpha-blended quad)
  m_pd3dDevice.SetRenderState(D3DRS_ALPHABLENDENABLE, iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_SRCBLEND,  D3DBLEND_SRCALPHA);
  m_pd3dDevice.SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA);

  m_pd3dDevice.SetStreamSource(0, m_pMirrorVB, 0, Sizeof(TMirrorVertex));
  m_pd3dDevice.SetFVF(TMirrorVertex_FVF);
  m_pd3dDevice.DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2);

  m_pd3dDevice.SetRenderState(D3DRS_ALPHABLENDENABLE, iFalse);

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.Render: HRESULT;
begin
  // Clear the viewport
  m_pd3dDevice.Clear(0, nil, D3DCLEAR_TARGET or D3DCLEAR_ZBUFFER, $000000ff, 1.0, 0);

  // Begin the scene
  if SUCCEEDED(m_pd3dDevice.BeginScene) then
  begin
    // Render the scene
    RenderScene;

    // Render the scene in the mirror
    RenderMirror;

    // End the scene.
    m_pd3dDevice.EndScene;
  end;

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.InitDeviceObjects:HRESULT;
var
  v: PAMirrorVertex;
begin
  // Initialize the font's internal textures
  m_pFont.InitDeviceObjects(m_pd3dDevice);

  // Set up the geometry objects
  if FAILED(m_pTeapot.Create_(m_pd3dDevice, PChar('Teapot.x'))) then
  begin
    Result:= D3DAPPERR_MEDIANOTFOUND;
    Exit;
  end;

  // Create a square for rendering the mirror
  if FAILED(m_pd3dDevice.CreateVertexBuffer(4*SizeOf(TMirrorVertex),
       D3DUSAGE_WRITEONLY, TMirrorVertex_FVF, D3DPOOL_MANAGED, m_pMirrorVB, nil)) then
  begin
    Result:= E_FAIL;
    Exit;
  end;

  // Initialize the mirror's vertices
  m_pMirrorVB.Lock(0, 0, Pointer(v), 0);
  v[0].p := D3DXVector3(-1.5, 1.5, 3.0);
  v[2].p := D3DXVector3(-1.5,-1.5, 3.0);
  v[1].p := D3DXVector3( 1.5, 1.5, 3.0);
  v[3].p := D3DXVector3( 1.5,-1.5, 3.0);
  v[0].n := D3DXVector3(0.0,0.0,-1.0);
  v[1].n := v[0].n;
  v[2].n := v[0].n;
  v[3].n := v[0].n;
  v[0].color := $80ffffff;
  v[1].color := $80ffffff;
  v[2].color := $80ffffff;
  v[3].color := $80ffffff;
  m_pMirrorVB.Unlock;

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.RestoreDeviceObjects: HRESULT;
var
  matWorld, matView, matProj: TD3DXMatrixA16;
  fAspect: Single;
  light: TD3DLight9;
begin
  m_pFont.RestoreDeviceObjects;

  // Set up the geometry objects
  m_pTeapot.RestoreDeviceObjects(m_pd3dDevice);

  // Set up the textures
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLOROP,   D3DTOP_MODULATE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_DIFFUSE);
  m_pd3dDevice.SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
  m_pd3dDevice.SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);

  // Set miscellaneous render states
  m_pd3dDevice.SetRenderState(D3DRS_DITHERENABLE,   iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_SPECULARENABLE, iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_ZENABLE,        iTrue);

  // Set up the matrices
  D3DXMatrixIdentity(matWorld);
  D3DXMatrixLookAtLH(matView, m_vEyePt, m_vLookatPt, m_vUpVec);
  fAspect := m_d3dsdBackBuffer.Width / m_d3dsdBackBuffer.Height;
  D3DXMatrixPerspectiveFovLH(matProj, D3DX_PI/4, fAspect, 1.0, 100.0);
  m_pd3dDevice.SetTransform(D3DTS_WORLD,      matWorld);
  m_pd3dDevice.SetTransform(D3DTS_VIEW,       matView);
  m_pd3dDevice.SetTransform(D3DTS_PROJECTION, matProj);

  // Set up a light
  if ((m_d3dCaps.VertexProcessingCaps and D3DVTXPCAPS_DIRECTIONALLIGHTS) = D3DVTXPCAPS_DIRECTIONALLIGHTS) or
     ((m_dwCreateFlags and D3DCREATE_HARDWARE_VERTEXPROCESSING) = 0) then
  begin
    D3DUtil_InitLight(light, D3DLIGHT_DIRECTIONAL, 0.2, -1.0, -0.2);
    m_pd3dDevice.SetLight(0, light);
    m_pd3dDevice.LightEnable(0, True);
  end;
  m_pd3dDevice.SetRenderState(D3DRS_AMBIENT, $ff555555);

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc:
//-----------------------------------------------------------------------------
function CMyD3DApplication.InvalidateDeviceObjects: HRESULT;
begin
  m_pFont.InvalidateDeviceObjects;
  m_pTeapot.InvalidateDeviceObjects;

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.DeleteDeviceObjects: HRESULT;
begin
  m_pFont.DeleteDeviceObjects;
  m_pTeapot.Destroy_;

  SAFE_RELEASE(m_pMirrorVB);

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device initialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
function CMyD3DApplication.ConfirmDevice(const pCaps: TD3DCaps9; dwBehavior: DWORD;
  adapterFormat: TD3DFormat; backBufferFormat: TD3DFormat): HRESULT;
begin
  Result:= E_FAIL;

  // GetTransform doesn't work on PUREDEVICE
  if (dwBehavior and D3DCREATE_PUREDEVICE) = D3DCREATE_PUREDEVICE then Exit;

  if ((dwBehavior and D3DCREATE_HARDWARE_VERTEXPROCESSING) = D3DCREATE_HARDWARE_VERTEXPROCESSING) or
     ((dwBehavior and D3DCREATE_MIXED_VERTEXPROCESSING) = D3DCREATE_MIXED_VERTEXPROCESSING) then
  begin
    if (pCaps.MaxUserClipPlanes < 4) then Exit;
  end;

  // Need to support post-pixel processing (for alpha blending)
  if FAILED(m_pD3D.CheckDeviceFormat(pCaps.AdapterOrdinal, pCaps.DeviceType,
       adapterFormat, D3DUSAGE_RENDERTARGET or D3DUSAGE_QUERY_POSTPIXELSHADER_BLENDING,
       D3DRTYPE_SURFACE, backBufferFormat))
  then Exit;

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FinalCleanup: HRESULT;
begin
  SAFE_DELETE(m_pFont);
  SAFE_DELETE(m_pTeapot);

  Result:= S_OK;
end;

end.

